from time import sleep
from typing import Any

from fastcore.utils import nested_idx
from termcolor import colored

from .auth import API_SCOPES, create_service

API_NAME = "slides"
API_VERSION = "v1"
PLACEHOLDER_TYPE = "BODY"
PLACEHOLDER_KEY = "pageElements"
BULLET_PRESET = "BULLET_DISC_CIRCLE_SQUARE"


def presentation_slides(service: Any, presentation_id: str) -> list[dict]:
    return (
        service.presentations()
        .get(presentationId=presentation_id)
        .execute()
        .get("slides", [])
    )


def find_placeholder_in_slide(
    slide: dict,
    placeholder_type: str = PLACEHOLDER_TYPE,
    keys: list[str] = [PLACEHOLDER_KEY],
) -> str:
    # page_elements = slide.get("pageElements")
    body_placeholder = ""
    page_elements = nested_idx(slide, *keys)
    if page_elements is None:
        return body_placeholder
    for pe in page_elements:
        if (
            pe.get("shape")
            and pe.get("shape").get("placeholder")
            and pe.get("shape").get("placeholder").get("type") == placeholder_type
        ):
            body_placeholder = pe.get("objectId")
            break

    return body_placeholder


def insert_text_request(object_id: str, text: str, insertion_index: int = 0) -> dict:
    return {
        "insertText": {
            "objectId": object_id,
            "insertionIndex": insertion_index,
            "text": text,
        }
    }


def create_bullet_request(
    object_id: str, text: str, bullet_preset: str = BULLET_PRESET
) -> dict:
    return {
        "createParagraphBullets": {
            "objectId": object_id,
            "textRange": {
                "type": "FIXED_RANGE",
                "startIndex": 0,
                "endIndex": len(
                    text
                ),  # Adjust end index to include the entire text for bullet
            },
            "bulletPreset": bullet_preset,  # Define the style of bullets you want to use
        }
    }


def add_text_to_slide(
    service: Any,
    presentation_id: str,
    text: str | list[str | list[str]] = ["HELLO", "DREAMAI"],
    placeholder_type: str | list[str] = ["TITLE", "BODY"],
    slide: dict | None = None,
    slide_id: int = -1,
) -> dict:
    assert slide or slide_id is not None, "No slide or slide_id provided."
    if not isinstance(text, list):
        text = [text]
    if not isinstance(placeholder_type, list):
        placeholder_type = [placeholder_type]
    slide = (
        slide
        or presentation_slides(presentation_id=presentation_id, service=service)[
            slide_id
        ]
    )
    batch_update_requests = []
    for text_content, ph_type in zip(text, placeholder_type):
        if ph_type == "NOTES":
            slide_keys = ["slideProperties", "notesPage", "pageElements"]
            ph_type = "BODY"
        else:
            slide_keys = ["pageElements"]
        placeholder = find_placeholder_in_slide(
            slide=slide, placeholder_type=ph_type, keys=slide_keys
        )
        if placeholder:
            # print(f"Found placeholder: {placeholder}")
            # print(f"Adding text: {text_content}")
            if ph_type in ["BODY", "NOTES"]:
                if not isinstance(text_content, list):
                    batch_update_requests.append(
                        insert_text_request(object_id=placeholder, text=text_content)
                    )
                    continue
                bullets = "\n".join(text_content)
                insertion_index = 0
                batch_update_requests.append(
                    insert_text_request(
                        object_id=placeholder,
                        text=bullets,
                        insertion_index=insertion_index,
                    )
                )
                bullet_request = create_bullet_request(
                    placeholder, "".join(text_content)
                )
                batch_update_requests.append(bullet_request)
            elif isinstance(text_content, str):
                batch_update_requests.append(
                    insert_text_request(object_id=placeholder, text=text_content)
                )
    # print(batch_update_requests)
    request_body = {"requests": batch_update_requests}
    return (
        service.presentations()
        .batchUpdate(presentationId=presentation_id, body=request_body)
        .execute()
    )


def create_presentation(
    service: Any | None = None,
    token_file: str = "",
    client_secrets_file: str = "",
    api_name: str = API_NAME,
    api_version: str = API_VERSION,
    scopes: list = API_SCOPES,
    name: str = "DREAMAI PRESENTATION",
    title: str = "DREAMAI",
    subtitle: str = "",
) -> str:
    assert (
        service is not None or token_file or client_secrets_file
    ), "No service or credentials provided."
    service = service or create_service(
        api_name=api_name,
        api_version=api_version,
        token_file=token_file,
        client_secrets_file=client_secrets_file,
        scopes=scopes,
    )
    body = {"title": name}
    presentation = service.presentations().create(body=body).execute()  # type: ignore
    presentation_id = presentation.get("presentationId")
    sleep(1)
    _ = add_text_to_slide(
        service=service,
        presentation_id=presentation_id,
        text=[title, subtitle],
        placeholder_type=["CENTERED_TITLE", "SUBTITLE"],
        slide_id=0,
    )
    print(colored(f"Created presentation with ID:{presentation_id}", "green"))
    return presentation_id


def add_slide(
    service: Any,
    presentation_id: str,
    title_text: str = "DREAMAI",
    body_text: str | list[str] = [
        "Generative AI",
        "Computer Vision",
        "Natural Language Processing",
    ],
    notes_text: str = "slide notes",
) -> dict:
    slides = presentation_slides(service=service, presentation_id=presentation_id)
    next_slide_id = max(1, len(slides))
    batch_update_requests = [
        {
            "createSlide": {
                "objectId": f"slide_{next_slide_id}",
                "insertionIndex": next_slide_id,
                "slideLayoutReference": {"predefinedLayout": "TITLE_AND_BODY"},
            }
        }
    ]
    body = {"requests": batch_update_requests}
    _ = (
        service.presentations()
        .batchUpdate(presentationId=presentation_id, body=body)
        .execute()
    )

    sleep(1)

    return add_text_to_slide(
        service=service,
        presentation_id=presentation_id,
        text=[title_text, body_text, notes_text],
        placeholder_type=["TITLE", "BODY", "NOTES"],
        slide_id=next_slide_id,
    )



import os
from typing import Any

from google.auth.transport.requests import Request
from google.oauth2.credentials import Credentials
from google_auth_oauthlib.flow import InstalledAppFlow
from googleapiclient.discovery import build

API_SCOPES = [
    "https://www.googleapis.com/auth/presentations",
    "https://www.googleapis.com/auth/drive.file",
]


def authenticate(
    token_file: str, client_secrets_file: str = "", scopes: str | list[str] = API_SCOPES
):
    creds = None
    if os.path.exists(token_file):
        creds = Credentials.from_authorized_user_file(
            filename=token_file, scopes=scopes
        )
    if not creds or not creds.valid:
        if creds and creds.expired and creds.refresh_token:
            creds.refresh(Request())
        elif client_secrets_file:
            flow = InstalledAppFlow.from_client_secrets_file(
                client_secrets_file=client_secrets_file,
                scopes=scopes,
            )
            creds = flow.run_local_server(port=0)
        else:
            raise Exception("No credentials found.")
        with open(token_file, "w") as token:
            token.write(creds.to_json())
    return creds


def create_service(
    api_name: str,
    api_version: str,
    token_file: str = "slides_tokens.json",
    client_secrets_file: str = "",
    scopes: list = API_SCOPES,
) -> Any:
    creds = authenticate(
        token_file=token_file,
        client_secrets_file=client_secrets_file,
        scopes=scopes,
    )
    service = build(serviceName=api_name, version=api_version, credentials=creds)
    return service



import os
from pathlib import Path

from .utils import deindent


def txt_to_prompt(prompt_file: str | Path) -> str:
    if not os.path.exists(prompt_file := str(prompt_file)):
        return ""
    with open(prompt_file, "r") as f:
        prompt = f.read()
    return deindent(prompt)


def process_prompt(prompt: str | Path | list[str]) -> str:
    if not prompt:
        return ""
    if isinstance(prompt, list):
        prompt = "\n\n---\n\n".join(list(prompt))
    elif isinstance(prompt, (Path, str)):
        prompt = str(prompt)
        if prompt.endswith(".txt"):
            prompt = txt_to_prompt(prompt)
    return deindent(str(prompt))


def titles_w_content_template(
    titles_dict: dict[str, str | Path | list[str]] = {},
    prefix: str = "",
    suffix: str = "",
) -> str:
    prompt = deindent(prefix)
    for title, content in titles_dict.items():
        if content:
            title = " ".join(title.split("_")).strip().title()
            content = process_prompt(content)
            if content:
                if len(prompt) > 0:
                    prompt += "\n\n"
                prompt += deindent(f"## {title} ##\n\n{deindent(content)}")
    prompt += "\n\n" + deindent(suffix)
    return deindent(prompt)



from chromadb import Collection as ChromaCollection
from langchain.text_splitter import RecursiveCharacterTextSplitter
from langchain_community.document_loaders import PyPDFLoader  # type: ignore
from langchain_core.documents import Document as LCDocument

from .chroma import (
    CHROMA_DEVICE,
    CHROMA_DIR,
    CHROMA_EMBEDDING_MODEL,
    chroma_collection,
    lc_docs_to_chroma_docs,
)

CHUNK_SIZE = 500
CHUNK_OVERLAP = 50
SEPARATORS = ["\n\n", "\n"]
CHROMA_COLLECTION_NAME = "pdf_collection"
CHROMA_DELETE_EXISTING = False


def pdf_to_docs(
    pdf_file: str,
    chunk_size: int = CHUNK_SIZE,
    chunk_overlap: int = CHUNK_OVERLAP,
    separators: list = SEPARATORS,
) -> list[LCDocument]:
    loader = PyPDFLoader(file_path=pdf_file)
    splitter = RecursiveCharacterTextSplitter(
        chunk_size=chunk_size,
        chunk_overlap=chunk_overlap,
        separators=separators,
        keep_separator=False,
    )
    docs = loader.load_and_split(splitter)
    return docs


def pdf_to_collection(
    pdf_file: str,
    chunk_size: int = CHUNK_SIZE,
    chunk_overlap: int = CHUNK_OVERLAP,
    separators: list[str] = SEPARATORS,
    collection_name: str = CHROMA_COLLECTION_NAME,
    persistent_dir: str = CHROMA_DIR,
    delete_existing: bool = CHROMA_DELETE_EXISTING,
    embedding_model: str = CHROMA_EMBEDDING_MODEL,
    device: str = CHROMA_DEVICE,
    add_links: bool = True,
) -> ChromaCollection:
    pdf_docs = pdf_to_docs(
        pdf_file=pdf_file,
        chunk_size=chunk_size,
        chunk_overlap=chunk_overlap,
        separators=separators,
    )
    collection = chroma_collection(
        name=collection_name,
        persistent_dir=persistent_dir,
        delete_existing=delete_existing,
        embedding_model=embedding_model,
        device=device,
    )
    collection.add(**lc_docs_to_chroma_docs(pdf_docs, add_links=add_links))
    return collection



from typing import Any

from termcolor import colored

from .auth import API_SCOPES, create_service

API_NAME = "forms"
API_VERSION = "v1"


def create_quiz(
    service: Any,
    token_file: str = "",
    client_secrets_file: str = "",
    api_name: str = API_NAME,
    api_version: str = API_VERSION,
    scopes: list = API_SCOPES,
    name: str = "DREAMAI QUIZ",
    title: str = "DREAMAI",
) -> dict:
    assert (
        service or token_file or client_secrets_file
    ), "No service or credentials provided."
    service = service or create_service(
        api_name=api_name,
        api_version=api_version,
        token_file=token_file,
        client_secrets_file=client_secrets_file,
        scopes=scopes,
    )
    new_form_request = {"info": {"title": title, "documentTitle": name}}
    update_request_body = {
        "requests": [
            {
                "updateSettings": {
                    "settings": {"quizSettings": {"isQuiz": True}},
                    "updateMask": "quizSettings.isQuiz",
                }
            }
        ]
    }
    form = service.forms().create(body=new_form_request).execute()
    quiz_id = form["formId"]
    _ = service.forms().batchUpdate(formId=quiz_id, body=update_request_body).execute()
    print(colored(f"Created quiz with ID:{quiz_id}", "green"))
    return quiz_id


def create_question_request(
    question: str, answers: list[str], correct_index: int, points_per_question: int = 1
) -> dict:
    return {
        "createItem": {
            "item": {
                "title": question,
                "questionItem": {
                    "question": {
                        "required": True,
                        "grading": {
                            "pointValue": points_per_question,
                            "correctAnswers": {
                                "answers": [{"value": answers[correct_index]}]
                            },
                            "whenRight": {"text": "You got it!"},
                            "whenWrong": {"text": "Sorry, that's wrong"},
                        },
                        "choiceQuestion": {
                            "type": "RADIO",
                            "options": [{"value": answer} for answer in answers],
                        },
                    }
                },
            },
            "location": {"index": 0},
        },
    }


def add_question(
    service: Any,
    quiz_id: str,
    question: str,
    answers: list[str],
    correct_index: int,
    points_per_question: int = 1,
):
    new_question_request = create_question_request(
        question=question,
        answers=answers,
        correct_index=correct_index,
        points_per_question=points_per_question,
    )
    request_body = {"requests": [new_question_request]}
    return service.forms().batchUpdate(formId=quiz_id, body=request_body).execute()



import json
from pathlib import Path
from typing import Callable
from uuid import uuid4

import chromadb
import torch
from chromadb import Collection as ChromaCollection
from chromadb.api.types import Include, QueryResult, GetResult
from chromadb.utils.embedding_functions import SentenceTransformerEmbeddingFunction
from langchain_core.documents import Document as LCDocument
from sentence_transformers import CrossEncoder
from termcolor import colored

CHROMA_EMBEDDING_MODEL = "multi-qa-mpnet-base-cos-v1"
CHROMA_DIR = "chroma_dir"
CHROMA_DEVICE = "cuda"
CHROMA_DELETE_EXISTING = False
CROSS_ENCODER_MODEL = "cross-encoder/ms-marco-MiniLM-L-6-v2"


def chroma_collection(
    name: str,
    persistent_dir: str = CHROMA_DIR,
    delete_existing: bool = CHROMA_DELETE_EXISTING,
    embedding_model: str = CHROMA_EMBEDDING_MODEL,
    device: str = CHROMA_DEVICE,
) -> ChromaCollection:
    chroma_client = chromadb.PersistentClient(path=persistent_dir)
    if delete_existing:
        try:
            chroma_client.delete_collection(name)
        except Exception as e:
            print(colored(f"Error deleting collection named {name}: {e}", "red"))
            pass
    embedding_function = SentenceTransformerEmbeddingFunction(
        model_name=embedding_model, device=device
    )
    collection = chroma_client.get_or_create_collection(
        name,
        embedding_function=embedding_function,  # type: ignore
    )
    return collection


def json_files_to_collection(
    flist: list[str | Path], chroma_collection: ChromaCollection
):
    ids = []
    docs = []
    metas = []
    for i, fname in enumerate(flist):
        with open(fname, "r") as f:
            data = json.load(f)
            ids.append(f"id_{uuid4()}")
            metas.append(
                {
                    "level": data["level"],
                    "type": data["type"],
                    "prev_id": "",
                    "next_id": "",
                }
            )
            docs.append(json.dumps(data))
    # print(colored(f"Adding {i} files to collection", "cyan"))
    chroma_collection.add(ids=ids, documents=docs, metadatas=metas)


def id_from_lc_doc(doc: LCDocument, doc_idx=None, **kwargs) -> str | None:
    try:
        file_name = doc.metadata.get("filename", doc.metadata.get("source"))
        doc_id = file_name or ""
        page_number = doc.metadata.get("page_number")
        if page_number is not None:
            doc_id += f"_page_{page_number}"
        if doc_idx is not None:
            doc_id += f"_{doc_idx}"
        return doc_id + f"_{uuid4()}"
    except Exception as e:
        print(colored(f"Error getting id from LCDocument: {e}", "red"))
        return None


def lc_docs_to_chroma_docs(
    docs: list[LCDocument],
    id_fn: Callable = id_from_lc_doc,
    add_links: bool = True,
) -> dict:
    chroma_ids = []
    chroma_docs = []
    chroma_metadatas = []
    for i, doc in enumerate(docs):
        chroma_ids.append(id_fn(doc=doc, doc_idx=i) or f"id_{i}")
        chroma_docs.append(doc.page_content)
    for i, doc in enumerate(docs):
        metadata = {
            k: v for k, v in doc.metadata.items() if type(v) in [str, int, float, bool]
        }
        metadata["prev_id"] = ""
        metadata["next_id"] = ""
        if add_links:
            metadata["prev_id"] = chroma_ids[i - 1] if i > 0 else ""
            metadata["next_id"] = chroma_ids[i + 1] if i < len(docs) - 1 else ""
        chroma_metadatas.append(metadata)
    return {"ids": chroma_ids, "documents": chroma_docs, "metadatas": chroma_metadatas}


def traverse_id(
    id: str,
    collection: ChromaCollection,
    direction: str = "prev",
    n_steps: int = 2,
) -> list:
    ids = []
    for _ in range(n_steps):
        metadata = collection.get(ids=[id])["metadatas"][0]  # type: ignore
        id = metadata.get(f"{direction}_id", "")  # type: ignore
        if not id:
            break
        ids.append(id)
    return ids


def traverse_ids(
    ids: str | list[str],
    collection: ChromaCollection,
    n_prev_links: int = 2,
    n_next_links: int = 2,
) -> list[list[str]]:
    if isinstance(ids, str):
        ids = [ids]
    res_ids = []
    for id in ids:
        prev_ids = traverse_id(
            id=id,
            collection=collection,
            direction="prev",
            n_steps=n_prev_links,
        )
        next_ids = traverse_id(
            id=id,
            collection=collection,
            direction="next",
            n_steps=n_next_links,
        )
        res_ids.append(prev_ids[::-1] + [id] + next_ids)
    return res_ids


def rerank_chroma_results(
    query_text: str,
    results: QueryResult,
    cross_encoder_model: str = CROSS_ENCODER_MODEL,
) -> dict:
    device = "cuda" if torch.cuda.is_available() else "cpu"
    cross_encoder = CrossEncoder(cross_encoder_model, device=device)
    pairs = [[query_text, doc] for doc in results["documents"][0]]  # type: ignore
    scores = cross_encoder.predict(pairs)
    scores_idx = sorted(
        range(len(scores)), key=lambda i: scores[i].item(), reverse=True
    )
    return {k: [[v[0][i] for i in scores_idx]] for k, v in results.items() if v}  # type: ignore


def query_collection(
    query_text: str,
    collection: ChromaCollection,
    n_results: int = 10,
    rerank_results: bool = False,
    n_prev_links: int = 2,
    n_next_links: int = 2,
    include: Include = ["metadatas", "documents"],
    reranker_model: str = CROSS_ENCODER_MODEL,
) -> tuple[list[GetResult], list[str]]:
    query_res = collection.query(
        query_texts=query_text, n_results=n_results, include=include
    )
    if rerank_results:
        query_res = rerank_chroma_results(
            query_text=query_text,
            results=query_res,
            cross_encoder_model=reranker_model,
        )
    init_ids = query_res["ids"][0]
    traversed_ids = traverse_ids(
        ids=init_ids,
        collection=collection,
        n_prev_links=n_prev_links,
        n_next_links=n_next_links,
    )
    results = [collection.get(ids=ids, include=include) for ids in traversed_ids]
    return results, init_ids
